﻿/* @(#) getwmp.c */
/********************************************************************/
/*                                                                  */
/* Program name: GETWMP                                             */
/*                                                                  */
/* Description: Sample C program that takes a user-provided file    */
/*              name and gets a message whose 'FileName' message    */
/*              property has a matching value. The message data     */
/*              payload is written to a file. If the message's      */
/*              'MsgGroupId' message property has a non-null value, */
/*              then that value is used to match the MD.GroupId of  */
/*              the next message to be retrieved, a message that is */
/*              (almost certainly) segmented.                       */
/*                                                                  */
/*                                                                  */
/********************************************************************/
/*                                                                  */
/* This program retrieves messages put by the 'putwmp' program. Be  */
/* sure to specify the names of the destination queue manager and   */
/* queue as arguments, thusly:                                      */
/*                            > getwmp.exe  QM2  AQ2                */
/*                                                                  */
/* Unlike the 'putwmp' program, the 'getwmp' program runs only in   */
/* 'fixed'solution mode. Note that there is no logical constraint   */
/* that compels getting the segmented message immediately after the */
/* reference message. It's done that way here because it's the most */
/* convenient. The solution pattern structure lends itself nicely   */
/* to asynchronously threading off the task of getting segmented    */
/* messages should anyone wish to try such an implementation.       */
/*                                                                  */
/********************************************************************/
/*                                                                  */
/*                                                                  */
/********************************************************************/
#include <io.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <windows.h>
#include "cmqc.h"                                               // standard IBM MQ include file

#define KIB                1024L
#define MIB                1048576L
#define OH                 1.15                                 // message buffer size over-head
#define MSGPROPCNT         4                                    // number of message properties
#define SELSTRMAXLEN       512                                  // max length of selector string

#define FILNAMEMAXLEN      64                                   // max length for a file name
#define MSGSIZEMAXLEN      8                                    // max character length for message size
#define GETMSGWAIT         5                                    // get message wait time-out (5 seconds)

#define sq(x)              (x * x)                              // x squared
#define abs(x)             ((x < 0) ? (x * -1) : x)             // |x|
#define MQVS_USE_VSLENGTH  0                                    // WMQv7 has foresaken this FRU


int main(int argc, char **argv)
{
  FILE *fsTextFile;
  errno_t Error;

  int   Quit                 = 0;
  int   ReturnValue          = 0;
  int   SysError             = 0;
  int   Index                = 0;

  int   QMgrConnected        = 0;
  int   MsgHndlDefined       = 0;
  int   ObjHndOneOpen        = 0;
  int   ObjHndTwoOpen        = 0;

  int   GetTheMessage        = 0;
  int   AllocateMemory       = 0;
  int   GetSgmntdMsg         = 0;
  int   QMgrNameLen;                                            // queue manager name length
  int   QueNameLen;                                             // queue name length

  enum { FileNameMPIdx,                                         // message properties array index #0 for file name
         MsgLengthMPIdx,                                        // message properties array index #1 for message size
         MsgGrpIdMPIdx,                                         // message properties array index #2 for GroupId
         MsgPropFillMPIdx };                                    // message properties array index #3 for 1KiB filler

  MQCHAR  SelectFileName[65] = { '\0' };                        // file name selector
  MQCHAR  FileNameValue[65]  = { '\0' };                        // file name message property value
  MQCHAR  MsgLengthValue[9]  = { '\0' };                        // message size message property value (up to "99999999\0")
  PMQCHAR PtrFileName        = (PMQCHAR) NULL;                  // pointer to file name
  PMQCHAR PtrHolder          = (PMQCHAR) NULL;                  // pointer holder

  MQCHAR  PropFillValue[KIB + 1]                = { '\0' };     // property filler message property value
  MQCHAR  QueObjSelStr[SELSTRMAXLEN]            = { '\0' };     // selector string
  MQCHAR  MsgGrpIdValue[MQ_GROUP_ID_LENGTH + 1] = { '\0' };     // segmented message group ID message property value

  MQLONG  FileNameMaxLen     = FILNAMEMAXLEN;
  MQLONG  MsgLengthMaxLen    = MSGSIZEMAXLEN;
  MQLONG  MsgGrpIdMaxLen     = MQ_GROUP_ID_LENGTH;

  MQMD    MsgHdrDesc         = {MQMD_DEFAULT};                  // message descriptor
  MQOD    QueObjDesc         = {MQOD_DEFAULT};                  // object descriptor
  MQGMO   GetMsgOpts         = {MQGMO_DEFAULT};                 // get message options
  MQLONG  QueOpenOpts        = MQCO_NONE;                       // MQOPEN options
  MQLONG  QueCloseOpts       = MQCO_NONE;                       // MQCLOSE options

  MQPD    MsgPrpDesc         = {MQPD_DEFAULT};                  // property descriptor
  MQCMHO  CrtMsgHndOpts      = {MQCMHO_DEFAULT};                // create message handle options
  MQDMHO  DltMsgHndOpts      = {MQDMHO_DEFAULT};                // delete message handle options
  MQIMPO  InqMsgPrpOpts      = {MQIMPO_DEFAULT};                // inquire message property options

  MQHCONN QMgrHandle;                                           // connection handle
  MQHOBJ  QueObjHnd1;                                           // object handle #1 (no selector)
  MQHOBJ  QueObjHnd2;                                           // object handle #2
  MQHMSG  MsgHandle;                                            // message handle
  MQLONG  CompCode;                                             // completion code
  MQLONG  Reason;                                               // reason code
  MQLONG  MsgBufSize;                                           // new message buffer size
  MQLONG  OldMsgBufSize;                                        // old message buffer size
  MQLONG  MsgLen;                                               // message length
  MQLONG  ThisMsgLen;                                           // current message length
  MQLONG  MsgPrpTyp;                                            // message property type

  PMQBYTE MsgBufPtr          = (PMQBYTE) NULL;                  // pointer to message buffer
  PMQCHAR QMgrName           = (PMQCHAR) NULL;                  // pointer to queue manager name
  PMQCHAR QueName            = (PMQCHAR) NULL;                  // pointer to queue name

  MQLONG  MsgPropCount       = MSGPROPCNT;
  MQLONG  BytesWritten       = 0;
  MQLONG  MinMsgBufSize      = 64 * KIB;                        // minimum message buffer size 64KiB

/********************************************************************************
**
** Four message properties are defined:
**   1) Name of the file contained in the message data payload
**   2) Size of the message data payload (a.k.a. file size)
**   3) Segmented message group ID
**   4) A 1KiB "filler" property
**
********************************************************************************/
  typedef struct tagMessageProperty stdMessageProperty;

  struct tagMessageProperty {
    MQCHARV Name;
    PMQCHAR Value;
    MQLONG  Length;
    MQLONG  MaxLength;
  };

  stdMessageProperty FileNameMP    = {                          // define message property #1
    { "FileName", 0, MQVS_USE_VSLENGTH, MQVS_NULL_TERMINATED, MQCCSI_APPL },
    FileNameValue,
    0,
    FileNameMaxLen
  };

  stdMessageProperty MsgLengthMP   = {                          // define message property #2
    { "MessageLength",  0,  MQVS_USE_VSLENGTH,  MQVS_NULL_TERMINATED,  MQCCSI_APPL },
    MsgLengthValue,
    0,
    MsgLengthMaxLen
  };

  stdMessageProperty MsgGrpIdMP    = {                          // define message property #3
    { "MessageGroupId", 0,  MQVS_USE_VSLENGTH,  MQVS_NULL_TERMINATED,  MQCCSI_APPL },
    MsgGrpIdValue,
    0,
    MsgGrpIdMaxLen
  };

  stdMessageProperty MsgPropFillMP = {                          // define message property #4
    { "PropertyFiller", 0,  MQVS_USE_VSLENGTH,  MQVS_NULL_TERMINATED,  MQCCSI_APPL },
    PropFillValue,
    0,
    KIB
  };

  stdMessageProperty * pstMessageProperty = (stdMessageProperty *) NULL;
  stdMessageProperty MsgPropsArray[MSGPROPCNT];                 // define array of message properties

  MsgPropsArray[FileNameMPIdx]    = FileNameMP;
  MsgPropsArray[MsgLengthMPIdx]   = MsgLengthMP;
  MsgPropsArray[MsgGrpIdMPIdx]    = MsgGrpIdMP;
  MsgPropsArray[MsgPropFillMPIdx] = MsgPropFillMP;

/********************************************************************************
**
** Initialize IBM MQ objects.
**
********************************************************************************/
  MsgHdrDesc.Version      = MQMD_CURRENT_VERSION;

  CrtMsgHndOpts.Version   = MQCMHO_CURRENT_VERSION;
  CrtMsgHndOpts.Options   = MQCMHO_DEFAULT_VALIDATION;          // *_VALIDATE; *_[DEFAULT | NO]_VALIDATION; *_NONE

  DltMsgHndOpts.Version   = MQDMHO_CURRENT_VERSION;
  DltMsgHndOpts.Options   = MQDMHO_NONE;

  MsgPrpTyp               = MQTYPE_STRING;                      // set message property type to 'string'  MQTYPE_BYTE_STRING

  QueObjDesc.Version      = MQOD_CURRENT_VERSION;
  QueObjDesc.ObjectType   = MQOT_Q;

  QueOpenOpts             = MQOO_INPUT_SHARED |
                            MQOO_FAIL_IF_QUIESCING;

  MsgPrpDesc.Version      = MQPD_CURRENT_VERSION;
  MsgPrpDesc.CopyOptions  = MQCOPY_NONE;                        // no copy options specified in property descriptor

  InqMsgPrpOpts.Version   = MQIMPO_CURRENT_VERSION;
  InqMsgPrpOpts.Options   = MQIMPO_CONVERT_VALUE |
                            MQIMPO_CONVERT_TYPE |
                            MQIMPO_INQ_FIRST;

  GetMsgOpts.Version      = MQGMO_CURRENT_VERSION;
  GetMsgOpts.MatchOptions = MQMO_NONE;
  GetMsgOpts.WaitInterval = GETMSGWAIT * 1000;                  // wait time in milliseconds

  GetMsgOpts.Options      = MQGMO_WAIT |                        // wait for a message
                            MQGMO_CONVERT |                     // convert to our CCSID
                            MQGMO_COMPLETE_MSG |                // we want the whole enchilada
                            MQGMO_NO_SYNCPOINT |                // no syncpoint for non-persistent messages
                            MQGMO_PROPERTIES_IN_HANDLE |        // we defined a message handle
                            MQPMO_FAIL_IF_QUIESCING;            // don't bother if QMgr is stopping

/********************************************************************************
**
** Check for queue manager and queue arguments.
**
********************************************************************************/
  if (argc < 3) {                                               // if: command line does not have required arguments
    printf("\nUSAGE: getwmp QMgrName QueueName\n");
    exit(1);
  }

  QMgrName    = argv[1];
  QueName     = argv[2];
  QMgrNameLen = strlen(argv[1]);
  QueNameLen  = strlen(argv[2]);

  if (QMgrNameLen > MQ_Q_MGR_NAME_LENGTH) {                     // if: supplied queue manager name length is too long
    printf("\nQueue manager \"%s\": name length = [%d] bytes, cannot exceed [%d] bytes. Try again.\n\n",
      QMgrName, QMgrNameLen, MQ_Q_MGR_NAME_LENGTH);
    exit(1);
  }

  if (QueNameLen > MQ_Q_NAME_LENGTH) {                          // if: supplied queue name length is too long
    printf("\nQueue \"%s\": name length = [%d] bytes, cannot exceed [%d] bytes. Try again.\n\n",
      QueName, QueNameLen, MQ_Q_MGR_NAME_LENGTH);
    exit(1);
  }

  printf("\nWelcome to \"Get Your Segmented Message\".\n");

  memcpy(QueObjDesc.ObjectName, QueName, (size_t) QueNameLen);  // copy queue name to object descriptor data structure
  MsgBufSize = 0;

/********************************************************************************
**
** Allocate message property name and value buffer memory.
**
********************************************************************************/
  for (Index = 0; Index < MsgPropCount; Index++) {              // for: all message properties
    if ((MsgPropsArray[Index].Value =                           // if: memory allocation failed
        (PMQCHAR) malloc(MsgPropsArray[Index].MaxLength * sizeof (MQCHAR))) == NULL) {

      SysError = GetLastError();
      printf("\nAllocate memory of [%d] bytes for MQ message property value buffer failed with error [%d].\n",
        MsgPropsArray[Index].MaxLength, SysError);
      exit(SysError);
    }

    else {                                                      // else: memory allocation successful
      printf("\nAllocation of [%d] byte memory buffer for message property \"%s\" succeeded.\n",
        MsgPropsArray[Index].MaxLength, (PMQCHAR) MsgPropsArray[Index].Name.VSPtr);
    }                                                           // end else: memory allocation successful
  }                                                             // end for: all message properties

/********************************************************************************
**
** Connect to queue manager.
**
********************************************************************************/
  MQCONN (QMgrName,                    // queue manager name
          &QMgrHandle,                 // connection handle
          &CompCode,                   // completion code
          &Reason);                    // reason code

  if (CompCode == MQCC_FAILED) {
    printf("\nMQCONN failed with error [%d].\\nn", Reason);
    ReturnValue = Reason;
    Quit = 1;
  }
  else {
    printf("\nConnect to queue manager \"%s\" succeeded.\n", QMgrName);
    QMgrConnected = 1;
  }

/********************************************************************************
**
** Create the message handle and assign to GMO.MsgHandle.
**
********************************************************************************/
  if (QMgrConnected) {
    MQCRTMH (QMgrHandle,               // connection handle
             &CrtMsgHndOpts,           // create message handle options
             &MsgHandle,               // message handle
             &CompCode,                // completion code
             &Reason);                 // reason code

    if (Reason != MQRC_NONE) {
      printf("\nCreate message handle failed with error [%d].\n\n", Reason);
      ReturnValue = Reason;
      Quit = 1;
    }
    else {
      printf("\nCreate message handle succeeded.\n");
      MsgHndlDefined = 1;
      GetMsgOpts.MsgHandle = MsgHandle;
    }
  }

/********************************************************************************
**
** Open QueObjHnd1: Get message from queue with no selector.
**
********************************************************************************/
  if (MsgHndlDefined) {
    QueObjDesc.SelectionString.VSPtr    = (MQPTR) NULL;
    QueObjDesc.SelectionString.VSLength = 0;

    MQOPEN (QMgrHandle,                // connection handle
            &QueObjDesc,               // object descriptor
            QueOpenOpts,               // object open options
            &QueObjHnd1,               // object handle #1
            &CompCode,                 // completion code
            &Reason);                  // reason code

    if (Reason != MQRC_NONE) {
      printf("\nOpen handle #1 on queue \"%s\" failed with error [%d].\n\n", QueName, Reason);
      ReturnValue = Reason;
      Quit = 1;
    }
    else {
      printf("\nOpen handle #1 on queue \"%s\" succeeded.\n", QueName);
      ObjHndOneOpen = 1;
    }
  }

/********************************************************************************
*********************************************************************************
**                >>>>>>> BEGIN MESSAGE REQUEST LOOP <<<<<<<
**
** Get file name pattern from user input.
**
*********************************************************************************
********************************************************************************/
  while (! Quit) {                                              // while: someone wants to enter file names
    printf("\n\nEnter full or partial name of file(s) to retrieve or 'Q' to quit: ");

    fgets(SelectFileName, sizeof(SelectFileName), stdin);
    PtrHolder = strchr(SelectFileName, '\n');                   // fgets includes NL which we must overwrite with null
    *PtrHolder = '\0';

    if (toupper(SelectFileName[0]) == 'Q') {                    // if: it's time to quit
      Quit = 1;
      continue;
    }                                                           // end if: it's time to quit

/********************************************************************************
**
** Build selector string.
**
** The file name provided by user input can constitute any contiguous part of
** the file's full name. This permits retrieval of multiple messages whose
** 'FileName' message property value matches the partial name provided.
**
********************************************************************************/
    sprintf_s(QueObjSelStr, sizeof(QueObjSelStr), "FileName LIKE '%%%s%%'", SelectFileName);

    QueObjDesc.SelectionString.VSPtr     = QueObjSelStr;
    QueObjDesc.SelectionString.VSOffset  = 0;
    QueObjDesc.SelectionString.VSBufSize = MQVS_USE_VSLENGTH;
    QueObjDesc.SelectionString.VSLength  = MQVS_NULL_TERMINATED;
    QueObjDesc.SelectionString.VSCCSID   = MQCCSI_APPL;

    printf("\nOpenning handle #2 on queue \"%s\" with selector string = \"%s\".\n",
      QueName, QueObjSelStr);

/********************************************************************************
**
** REMEMBER: When the selector string is changed, the relevant queue must be
**           closed and re-opened with the new selector.
**
** Open QueObjHnd2: Get message from queue with assigned selector.
**
********************************************************************************/
    MQOPEN (QMgrHandle,                // connection handle
            &QueObjDesc,               // object descriptor
            QueOpenOpts,               // object open options
            &QueObjHnd2,               // object handle #2
            &CompCode,                 // completion code
            &Reason);                  // reason code

    if (Reason != MQRC_NONE) {
      printf("\nOpen handle #2 on queue \"%s\" failed with error [%d].\n\n", QueName, Reason);

      ReturnValue = Reason;
      Quit = 1;
      continue;
    }
    else {
      printf("\nOpen handle #2 on queue \"%s\" succeeded.\n", QueName);

      ObjHndTwoOpen  = 1;
      GetTheMessage  = 1;                                       // let's get a message
      AllocateMemory = 0;                                       // don't know yet if we need any memory
      GetSgmntdMsg   = 0;                                       // the next MQGET will notbe for a segmented message
    }

/********************************************************************************
*********************************************************************************
**                >>>>>>> BEGIN MESSAGE RETRIEVAL LOOP <<<<<<<
**
** Get message(s) from queue.
**
*********************************************************************************
********************************************************************************/
    while (GetTheMessage) {                                     // while: we're tasked to get a message

/********************************************************************************
**
** If requested, allocate additional memory for message buffer.
**
********************************************************************************/
      if (AllocateMemory == 1) {                                // if: additional memory was requested
        AllocateMemory = 0;
        SysError = 0;
        ThisMsgLen = (MQLONG) (ThisMsgLen * OH);                // set new message buffer size 15% bigger than requested

        OldMsgBufSize = MsgBufSize;
        MsgBufSize =                                            // enforce minimum message buffer size
          (ThisMsgLen > MinMsgBufSize) ? ThisMsgLen : MinMsgBufSize;

        if (OldMsgBufSize == 0L) {                              // if: this is the initial memory allocation request
          if ((MsgBufPtr = (PMQBYTE) malloc(MsgBufSize * sizeof(MQBYTE))) == NULL) {   // if: allocate memory failed
            SysError = GetLastError();
          }
        }                                                       // end if: this is the initial memory allocation request

        else {                                                  // else: this is a memory reallocation request
          if ((MsgBufPtr = (PMQBYTE) realloc(MsgBufPtr, MsgBufSize * sizeof(MQBYTE))) == NULL) { // if: reallocate memory failed
            SysError = GetLastError();
          }
        }                                                       // end else: this is a memory reallocation request

        if (SysError) {                                         // if: memory allocation failed
          printf("\n   Attempt to allocate [%d] bytes of memory for message buffer failed with error [%d].\n\n",
            MsgBufSize, SysError);

          ReturnValue = SysError;
          SysError = 0;
          GetTheMessage = 0;
          Quit = 1;
          continue;
        }                                                       // end if: memory allocation failed

        printf("\n   Allocate [%d] bytes of memory for message buffer succeeded.\n", MsgBufSize);
      }                                                         // end if: additional memory was requested

      if (GetSgmntdMsg == 1) {                                  // if: we're getting a segmented message
        printf("\n   Retrieving [%s] byte segmented message matching GroupId = \"%s\".\n",
          MsgPropsArray[MsgLengthMPIdx].Value, MsgPropsArray[MsgGrpIdMPIdx].Value);
      }
      else {
        printf("\n   Retrieving message using selector \"%s\", timeout = [%d] seconds.\n", QueObjSelStr, GETMSGWAIT);
      }

      MQGET (QMgrHandle,                                        // connection handle
             (GetSgmntdMsg == 1) ? QueObjHnd1 : QueObjHnd2,     // object handle
             (GetSgmntdMsg == 1) ? &MsgHdrDesc : (PMQMD) NULL,  // message header
             &GetMsgOpts,                                       // get message options
             MsgBufSize,                                        // message buffer size
             MsgBufPtr,                                         // message buffer
             &MsgLen,                                           // message length
             &CompCode,                                         // completion code
             &Reason);                                          // reason code

      if (Reason == MQRC_NONE) {
        if (MsgLen > 1)
          printf("\nMQGET of [%d] byte message from queue \"%s\" succeeded.\n", MsgLen, QueName);
      }
      else {                                                    // else: MQGET failed
        switch (Reason) {                                       // switch: (Reason)
          case MQRC_NO_MSG_AVAILABLE:
            printf("\nNo (more) messages available using selector \"%s\".\n", QueObjSelStr);

/********************************************************************************
**
** There are no more messages to be retrieved using the current selector string.
** Close queue object #2 it it's open.
**
********************************************************************************/
            if (ObjHndTwoOpen) {                                // if: object handle #2 is open

              MQCLOSE (QMgrHandle,     // connection handle
                       &QueObjHnd2,    // object handle
                       QueCloseOpts,   // object close options
                       &CompCode,      // completion code
                       &Reason);       // reason code

              if (Reason != MQRC_NONE) {
                printf("\nClose handle #2 on queue \"%s\" failed with error [%d].\n\n",
                  QueName, Reason);

                ReturnValue = Reason;
                GetTheMessage = 0;
                Quit = 1;
                break;
              }
              else {
                printf("\nClose handle #2 on queue \"%s\" succeeded.\n", QueName);
                ObjHndTwoOpen = 0;
              }
            }                                                   // end if: object handle #2 is open

            GetTheMessage = 0;
            break;

/********************************************************************************
**
** If message is truncated, allocate more memory and MQGET again.
**
********************************************************************************/
          case MQRC_TRUNCATED_MSG_FAILED:                       // truncated message
            printf("\nCurrent message buffer size of [%d] bytes is too small to retrieve [%d] byte message from \
queue \"%s\".", MsgBufSize, MsgLen, QueName);

            AllocateMemory = 1;                                 // request allocation of more memory
            ThisMsgLen = MsgLen;                                // to handle this message's data payload
            break;

          default:
            printf("\nMQGET of [%d] byte message from queue \"%s\" failed with error [%d].\n",
              MsgLen, QueName, Reason);

            ReturnValue = Reason;
            GetTheMessage = 0;
            Quit = 1;
            break;
        }                                                       // end switch: (Reason)
      }                                                         // end else: MQGET failed

      if (AllocateMemory == 1) continue;                        // allocate additional memory and retry the MQGET
      if (GetTheMessage == 0) break;                            // MQGET us no more

/********************************************************************************
**
** If this message is not segmented, then inquire the message's properties.
**
********************************************************************************/
      if (GetSgmntdMsg != 1) {                                  // if: this is not a segmented message
        for (Index = 0; Index < MsgPropCount; Index++) {        // for: inquire all message properties
          MQINQMP (QMgrHandle,                                  // connection handle
                   MsgHandle,                                   // message handle
                   &InqMsgPrpOpts,                              // message property inquiry options
                   &MsgPropsArray[Index].Name,                  // property name to be inquired
                   &MsgPrpDesc,                                 // property descriptor
                   &MsgPrpTyp,                                  // property type
                   MsgPropsArray[Index].MaxLength,              // size of the property value buffer
                   MsgPropsArray[Index].Value,                  // property value buffer
                   &MsgPropsArray[Index].Length,                // size of the returned property value
                   &CompCode,                                   // completion code
                   &Reason);                                    // reason code

          if (Reason == MQRC_NONE) {
            *(MsgPropsArray[Index].Value + MsgPropsArray[Index].Length) = '\0';     // NULL terminate the MP value
          }
          else {
            printf("\nInquiry of message property name \"%s\" failed with error [%d].\n",
              (PMQCHAR) MsgPropsArray[Index].Name.VSPtr, Reason);

            ReturnValue = Reason;
            GetTheMessage = 0;
            Quit = 1;
            continue;
          }
        }                                                       // end for: inquire all message properties

/********************************************************************************
**
** Get the size of the message data payload (MDP).
**
********************************************************************************/
        if ((ThisMsgLen = atoi(MsgPropsArray[MsgLengthMPIdx].Value)) == 0) {
          SysError = GetLastError();

          printf("\nInteger conversion of message property [%d] failed with error [%d]:\n\
   Name = \"%s\", Value = \"%s\".\n\n", MsgLengthMPIdx + 1, SysError,
            (PMQCHAR) MsgPropsArray[MsgLengthMPIdx].Name.VSPtr, MsgPropsArray[MsgLengthMPIdx].Value);

          ReturnValue = SysError;
          GetTheMessage = 0;
          Quit = 1;
          continue;
        }

        if (MsgBufSize < (MQLONG) (ThisMsgLen * OH)) {          // if: current message buffer size < MDP + 15%
          printf("\nCurrent message buffer size of [%d] bytes is too small to retrieve [%d] byte message from \
queue \"%s\".\n", MsgBufSize, ThisMsgLen, QueName);

          AllocateMemory = 1;                                   // request allocation of additional memory
        }

/********************************************************************************
**
** Assign pointer to the 'FileName' message property value.
**
********************************************************************************/
        PtrFileName = MsgPropsArray[FileNameMPIdx].Value;

/********************************************************************************
**
** If 'MessageGroupId' message property value is not NULL, then we have a
** reference message for a segmented message.
**
********************************************************************************/
        if (MsgPropsArray[MsgGrpIdMPIdx].Length > 0) {          // if: we have a segmented message to retrieve
          printf("\n   Retrieved reference message for [%d] byte segmented message.\n", ThisMsgLen);

          GetSgmntdMsg = 1;                                     // set the get-segmented-message flag

/********************************************************************************
**
** Set up MQGET for retrieving a message by matching on the MD.GroupId attribute
** value (that we get from the 'MessageGroupId' message property supplied by this
** reference message).
**
** NOTE: The reference message for a segmented message does not itself have a
**       message data payload which is why we immediately proceed to get the
**       segmented message.
**
********************************************************************************/
          memcpy(MsgHdrDesc.GroupId,                            // copy message group ID to MD.GroupId
                 MsgPropsArray[MsgGrpIdMPIdx].Value,
                 MsgPropsArray[MsgGrpIdMPIdx].Length);

          GetMsgOpts.MatchOptions = MQMO_MATCH_GROUP_ID;        // get the message that matches GroupId
          GetMsgOpts.Options |= MQGMO_NO_PROPERTIES;            // no message properties

          if (GetMsgOpts.Options & MQGMO_PROPERTIES_IN_HANDLE)  // unset MQGMO_PROPERTIES_IN_HANDLE
            GetMsgOpts.Options ^= MQGMO_PROPERTIES_IN_HANDLE;

          continue;                                             // proceed to get the segmented message
        }                                                       // end if: we have a segmented message to retrieve
      }                                                         // end if: this is not a segmented message

      else {                                                    // else: this message was segmented
        GetSgmntdMsg = 0;                                       // the next MQGET will not be for a segmented message

/********************************************************************************
**
** Set up MQGET for retrieving the next message which will not be segmented but
** will have message properties defined on it.
**
********************************************************************************/
        memcpy(MsgHdrDesc.GroupId,                              // copy MQGI_NONE to MD.GroupId
               MQGI_NONE,
               MQ_GROUP_ID_LENGTH);

        GetMsgOpts.MatchOptions = MQMO_NONE;                    // no GroupId match options
        GetMsgOpts.Options |= MQGMO_PROPERTIES_IN_HANDLE;       // return message properties in messsage handle

        if (GetMsgOpts.Options & MQGMO_NO_PROPERTIES)           // unset MQGMO_NO_PROPERTIES
          GetMsgOpts.Options ^= MQGMO_NO_PROPERTIES;
      }                                                         // end else: this message was segmented

/********************************************************************************
*********************************************************************************
**             >>>>>>> WRITE MESSAGE DATA PAYLOAD TO A FILE <<<<<<<
**
** Open file for write.
**
*********************************************************************************
********************************************************************************/
      if (((Error = fopen_s(&fsTextFile, PtrFileName, "wb")) != 0) ||
          (fsTextFile == NULL)) {                               // if: file open failed

        SysError = GetLastError();
        if (fsTextFile != NULL) clearerr(fsTextFile);

        printf("\nOpen file \"%s\" for write failed with error [%d], syserror [%d].\n",
          PtrFileName, (int) Error, SysError);

        ReturnValue = SysError;
        GetTheMessage = 0;
        Quit = 1;
        continue;
      }                                                         // end if: file open failed

      else {                                                    // else: file open succeeded
        printf("\nOpen file \"%s\" for write succeeded.\n", PtrFileName);

/********************************************************************************
**
**  Write contents of message buffer to file.
**
********************************************************************************/
        BytesWritten = fwrite(MsgBufPtr, sizeof(MQBYTE), MsgLen, fsTextFile);

        if (BytesWritten != MsgLen) {                           // if: write to file failed
          SysError = GetLastError();
          if (fsTextFile != NULL) clearerr(fsTextFile);

          printf("\nWrite to file \"%s\" failed with error [%d].\n", PtrFileName, SysError);

          ReturnValue = SysError;
          GetTheMessage = 0;
          Quit = 1;
        }                                                       // end if: write to file failed

        printf("\nWrite of [%d] bytes to file \"%s\" succeeded.\n", BytesWritten, PtrFileName);

/********************************************************************************
**
**  Close the file.
**
********************************************************************************/
        if (fclose(fsTextFile) == EOF) {                        // if: file close failed
          SysError = GetLastError();
          if (fsTextFile != NULL) clearerr(fsTextFile);

          printf("\nAttempt to close file \"%s\" failed with error [%d].\n", PtrFileName, SysError);

          ReturnValue = SysError;
          GetTheMessage = 0;
          Quit = 1;
        }                                                       // end if: file close failed

        else {                                                  // else: file close succeeded
          printf("\nClose file \"%s\" succeeded.\n", PtrFileName);
        }
      }                                                         // end else: file close succeeded
    }                                                           // end while: we're tasked to get a message
  }                                                             // end while: someone wants to enter file names

/********************************************************************************
**
** Close queue objects, delete message handle and close queue manager connection.
**
********************************************************************************/
  if (ObjHndOneOpen) {                                          // if: queue object #1 open
    MQCLOSE (QMgrHandle,               // connection handle
             &QueObjHnd1,              // object handle
             QueCloseOpts,             // object close options
             &CompCode,                // completion code
             &Reason);                 // reason code

    if (Reason != MQRC_NONE) {
      printf("\nClose handle #1 on queue \"%s\" failed with error [%d].\n\n", QueName, Reason);
      ReturnValue = Reason;
    }
    else {
      printf("\nClose handle #1 on queue \"%s\" succeeded.\n", QueName);
      ObjHndOneOpen = 0;
    }
  }                                                             // end if: queue object #1 open

  if (ObjHndTwoOpen) {                                          // if: queue object #2 open
    MQCLOSE (QMgrHandle,               // connection handle
             &QueObjHnd2,              // object handle
             QueCloseOpts,             // object close options
             &CompCode,                // completion code
             &Reason);                 // reason code

    if (Reason != MQRC_NONE) {
      printf("\nClose handle #2 on queue \"%s\" failed with error [%d].\n\n", QueName, Reason);
      ReturnValue = Reason;
    }
    else {
      printf("\nClose handle #2 on queue \"%s\" succeeded.\n", QueName);
      ObjHndTwoOpen = 0;
    }
  }                                                             // end if: queue object #2 open

  if (MsgHndlDefined) {                                         // if: message handle defined
    MQDLTMH (QMgrHandle,               // connection handle
             &MsgHandle,               // message handle
             &DltMsgHndOpts,           // delete message handle options
             &CompCode,                // completion code
             &Reason);                 // reason code

    if (CompCode == MQCC_FAILED) {
      printf("\nDelete message handle failed with error [%d].\n\n", Reason);
      ReturnValue = Reason;
    }
    else {
      printf("\nMessage handle delete succeeded.\n");
      MsgHndlDefined = 0;
    }
  }                                                             // end if: message handle defined

  if (QMgrConnected) {                                          // if: queue manager is connected
    MQDISC (&QMgrHandle,
            &CompCode,
            &Reason);

    if (CompCode == MQCC_FAILED) {
      printf("\nDisconnect from queue manager \"%s\" failed with error [%d].\n\n", QMgrName, Reason);
      ReturnValue = Reason;
    }
    else {
      printf("\nDisconnect from queue manager \"%s\" succeeded.\n", QMgrName);
      QMgrConnected = 0;
    }
  }                                                             // end if: queue manager is connected

  printf("\nCome back soon. Bye!\n\n");
  exit(ReturnValue);
}
