﻿/* @(#) putwmp.c */
/********************************************************************/
/*                                                                  */
/* Program name: PUTWMP                                             */
/*                                                                  */
/* Description: Sample C program that sets properties of a message  */
/*              handle using MQSETMP. A user-provided file name is  */
/*              used to access the file and read its contents into  */
/*              the message's data payload. The message is put to   */
/*              a remote queue to demonstrate (among other things)  */
/*              auto-segmentation of messages by the queue manager. */
/*                                                                  */
/********************************************************************/
/*                                                                  */
/* This program is designed to emulate the conditions that existed  */
/* when the failure scenario was discovered and subsequently when   */
/* the solution was implemented.                                    */
/*                                                                  */
/* Run mode '0' emulates the actual original messaging application  */
/* which specified a MQMD for the message descriptor argument to    */
/* the MQPUT call. The salient feature of this run mode is the      */
/* demonstration of a successful MQ completion code (MQCC) returned */
/* to the application after the MQPUT of a segmented message whose  */
/* first segment is too large, preventing retrieval by the down-    */
/* stream MCA and resulting with the queue manager stopping the     */
/* channel.                                                         */
/*                                                                  */
/* Run mode '1' is the same environment as '0' except the message   */
/* descriptor argument passed to the MQPUT call is (PMQMD) NULL.    */
/* This run mode demonstrates the return of MQ reason codes 2030 or */
/* 2031 to the application. Message segmentation still fails but    */
/* this time the offending message is not put onto the transmission */
/* queue thus sparing the channel.                                  */
/*                                                                  */
/* Run mode '2' emulates the 'fixed' MQ solution whereby a message  */
/* that is to be segmented is preceeded by a reference message with */
/* solution-specific message properties "MessageSize" and "GroupId" */
/* defined on it (in addition to any application-specific message   */
/* properties). The segmented message itself has no properties      */
/* defined on it so message retrieval by the getting application is */
/* done by setting the MQGMO match option to MQMO_MATCH_GROUP_ID.   */
/*                                                                  */
/* After the required three command line arguments (queue manager,  */
/* queue, runmode) you can specify an optional fourth command line  */
/* argument, 'MaxMessageSize'. The default value is 4MiB (4194304)  */
/* bytes.                                                           */
/*                                                                  */
/* A fifth optional argument, 'MaxBytesToRead', can be provided to  */
/* specify the number of bytes to be read from the text input file. */
/*                                                                  */
/* A sixth optional argument, 'MsgLenMod', can be provided to       */
/* reduce a given message's actual length processed by the MQPUT by */
/* the number of bytes specified.                                   */
/*                                                                  */
/*                                                                  */
/* Note that the command line arguments are positional. If you want */
/* to provide a value for max number of bytes to read but you want  */
/* to accept the default value for max message length, specify '0'  */
/* for the 'MaxMessageLength' argument. Ditto for the 'MsgLenMod'   */
/* argument.                                                        */
/*                                                                  */
/********************************************************************/
/*                                                                  */
/*                                                                  */
/********************************************************************/
#include <io.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>

#include <windows.h>
#include "cmqc.h"                                               // standard WMQ include file

#define NOT                !
#define KIB                1024L
#define MIB                1048576L
#define CLEARMQMD          MQ_MSG_HEADER_LENGTH
#define CLEARMPDS          58                                   // size of additional message property data
#define CLEARXMIT          428                                  // size of additional header data
#define MAXMSGSIZE         (MIB * 4)                            // max message size is 4MiB
#define MAXFILESIZE        (MIB * 10)                           // max text file size is 10 MiB
#define GRPIDHEADER        "SegMsgGroupIdHdr"                   // segmented message group ID
#define RNDNUMSIZE         8                                    // random number size
#define MSGPROPCNT         4                                    // number of message properties

#define FILNAMEMAXLEN      64
#define MSGSIZEMAXLEN      8

#define sq(x)              (x * x)                              // x squared
#define abs(x)             ((x < 0) ? (x * -1) : x)             // |x|
#define MQVS_USE_VSLENGTH  0                                    // WMQv7 has foresaken this

int main(int argc, char **argv)
{
  FILE *fsTextFile;
  errno_t Error;
  struct _stat FileStatBuf;

  int   Result               = 0;
  int   SysError             = 0;
  int   Index                = 0;
  int   argvLen              = 0;
  int   argCount             = 0;
  int   MsgLenMod            = 0;
  int   ReadMaxBytes         = 0;
  int   optArgs[6]           = { 0 };

  int   RunMode              = 0;                               // run broken (0) or fixed (1) mode
  int   BigMsgFlag           = 0;                               // big message flag
  int   GrpMsgNum            = 1;                               // big message group message number (1 or 2)
  int   MQPutSwitch;                                            // MQMD v. MsgHndl in MQPUT
  int   QMgrNameLen;                                            // queue manager name length
  int   QueNameLen;                                             // queue name length

  enum { FileNameMPIdx,                                         // message properties array index #0 for file name
         MsgLengthMPIdx,                                        // message properties array index #1 for message size
         MsgGrpIdMPIdx,                                         // message properties array index #2 for GroupId
         MsgPropFillMPIdx };                                    // message properties array index #3 for 1KiB filler

  MQCHAR  TextFileName[256]  = { '\0' };                        // text file FQN
  MQCHAR  FileNameValue[65]  = { '\0' };                        // file name message property value
  MQCHAR  RandID[9]          = { '\0' };                        // character form of random number
  MQCHAR  MsgLengthValue[9]  = { '\0' };                        // message size message property value (up to "99999999\0")
  PMQCHAR PtrFileName        = (PMQCHAR) NULL;                  // pointer to text file name
  PMQCHAR PtrHolder          = (PMQCHAR) NULL;                  // pointer holder

  MQCHAR  PropFillValue[KIB + 1]                = { '\0' };     // property filler message property value
  MQCHAR  MsgGrpIdValue[MQ_GROUP_ID_LENGTH + 1] = { '\0' };     // segmented message group ID message property value

  MQLONG  FileNameMaxLen     = FILNAMEMAXLEN;
  MQLONG  MsgLengthMaxLen    = MSGSIZEMAXLEN;
  MQLONG  MsgGrpIdMaxLen     = MQ_GROUP_ID_LENGTH;

  MQMD    MsgHdrDesc         = {MQMD_DEFAULT};                  // message descriptor
  MQOD    QueObjDesc         = {MQOD_DEFAULT};                  // object descriptor
  MQPMO   PutMsgOpts         = {MQPMO_DEFAULT};                 // put message options
  MQLONG  QueOpenOpts        = MQCO_NONE;                       // MQOPEN options
  MQLONG  QueCloseOpts       = MQCO_NONE;                       // MQCLOSE options

  MQPD    MsgPrpDesc         = {MQPD_DEFAULT};                  // property descriptor
  MQCMHO  CrtMsgHndOpts      = {MQCMHO_DEFAULT};                // create message handle options
  MQDMHO  DltMsgHndOpts      = {MQDMHO_DEFAULT};                // delete message handle options
  MQSMPO  SetMsgPrpOpts      = {MQSMPO_DEFAULT};                // set message property options

  MQHCONN QMgrHandle;                                           // connection handle
  MQHMSG  MsgHandle;                                            // message handle
  MQLONG  CompCode;                                             // completion code
  MQLONG  Reason;                                               // reason code
  MQLONG  MsgBufSize;                                           // new message buffer size
  MQLONG  OldMsgBufSize;                                        // old message buffer size
  MQLONG  MsgLen;                                               // message length
  MQLONG  TrueMsgLen;                                           // message length + size of all message properties + clearance
  MQLONG  MsgPrpTyp;                                            // message property type

  PMQBYTE MsgBufPtr          = (PMQBYTE) NULL;                  // pointer to message buffer
  PMQCHAR QMgrName           = (PMQCHAR) NULL;                  // pointer to queue manager name
  PMQCHAR QueName            = (PMQCHAR) NULL;                  // pointer to queue name

  MQLONG  MaxMsgSize         = MAXMSGSIZE;
  MQLONG  MaxFileSize        = MAXFILESIZE;
  MQLONG  TextFileSize       = 0;
  MQLONG  BytesRead          = 0;

  MQLONG  MsgPropCount       = MSGPROPCNT;
  MQLONG  MsgPropSize        = 0;
  MQLONG  MsgPropsInitSize   = 0;
  MQLONG  MsgPropsTotalSize  = 0;

/********************************************************************************
**
** Four message properties are defined:
**   1) Name of the file contained in the message data payload
**   2) Size of the message data payload (a.k.a. file size)
**   3) Segmented message group ID
**   4) A 1KiB "filler" property
**
********************************************************************************/
  typedef struct tagMessageProperty stdMessageProperty;

  struct tagMessageProperty {
    MQCHARV Name;
    PMQCHAR Value;
    MQLONG  Length;
    MQLONG  MaxLength;
  };

  stdMessageProperty FileNameMP    = {                          // define message property #1
    { "FileName", 0, MQVS_USE_VSLENGTH, MQVS_NULL_TERMINATED, MQCCSI_APPL },
    FileNameValue,
    0,
    FileNameMaxLen
  };

  stdMessageProperty MsgLengthMP   = {                          // define message property #2
    { "MessageLength",  0,  MQVS_USE_VSLENGTH,  MQVS_NULL_TERMINATED,  MQCCSI_APPL },
    MsgLengthValue,
    0,
    MsgLengthMaxLen
  };

  stdMessageProperty MsgGrpIdMP    = {                          // define message property #3
    { "MessageGroupId", 0,  MQVS_USE_VSLENGTH,  MQVS_NULL_TERMINATED,  MQCCSI_APPL },
    MsgGrpIdValue,
    0,
    MsgGrpIdMaxLen
  };

  stdMessageProperty MsgPropFillMP = {                          // define message property #4
    { "PropertyFiller", 0,  MQVS_USE_VSLENGTH,  MQVS_NULL_TERMINATED,  MQCCSI_APPL },
    PropFillValue,
    0,
    KIB
  };

  stdMessageProperty * pstMessageProperty = (stdMessageProperty *) NULL;
  stdMessageProperty MsgPropsArray[MSGPROPCNT];

  MsgPropsArray[FileNameMPIdx]    = FileNameMP;
  MsgPropsArray[MsgLengthMPIdx]   = MsgLengthMP;
  MsgPropsArray[MsgGrpIdMPIdx]    = MsgGrpIdMP;
  MsgPropsArray[MsgPropFillMPIdx] = MsgPropFillMP;

  MsgPropSize = sizeof (MQCHARV) + sizeof (PMQCHAR) + (2 * sizeof (MQLONG));
  MsgPropsInitSize = MsgPropCount * MsgPropSize;

  for (Index = 0; Index < MsgPropCount; Index++) {
    MsgPropsInitSize += strlen(MsgPropsArray[Index].Name.VSPtr);
  }

/********************************************************************************
**
** Initialize IBM MQ objects.
**
********************************************************************************/
  MsgHdrDesc.Version      = MQMD_CURRENT_VERSION;

  CrtMsgHndOpts.Version   = MQCMHO_CURRENT_VERSION;
  CrtMsgHndOpts.Options   = MQCMHO_DEFAULT_VALIDATION;          // *_VALIDATE; *_[DEFAULT | NO]_VALIDATION; *_NONE

  DltMsgHndOpts.Version   = MQDMHO_CURRENT_VERSION;
  DltMsgHndOpts.Options   = MQDMHO_NONE;

  SetMsgPrpOpts.Version   = MQSMPO_CURRENT_VERSION;
  SetMsgPrpOpts.Options   = MQSMPO_SET_FIRST;                   // set message prop value for 1st matched name or create if none found

  MsgPrpTyp               = MQTYPE_STRING;                      // set message property type to 'string'  MQTYPE_BYTE_STRING

  QueObjDesc.Version      = MQOD_CURRENT_VERSION;
  QueObjDesc.ObjectType   = MQOT_Q;

  QueOpenOpts             = MQOO_OUTPUT |
                            MQOO_FAIL_IF_QUIESCING;

  MsgPrpDesc.Version      = MQPD_CURRENT_VERSION;
  MsgPrpDesc.CopyOptions  = MQCOPY_NONE;                        // no copy options specified in property descriptor

  PutMsgOpts.Version      = MQPMO_CURRENT_VERSION;
  PutMsgOpts.Action       = MQACTP_NEW;                         // new message is always being put
  PutMsgOpts.Options      = MQPMO_NEW_MSG_ID |                  // create a new message ID for every message
                            MQPMO_NEW_CORREL_ID |               // create a new correlation  ID for every message
                            MQPMO_SYNC_RESPONSE |               // wait for MQPUT1 return
                            MQPMO_NO_SYNCPOINT |                // no syncpoint for non-persistent messages
                            MQPMO_FAIL_IF_QUIESCING;            // don't bother if QMgr is stopping

/********************************************************************************
**
** Initialize the PropertyFiller message property.
**
********************************************************************************/
  for (Index = 0; Index < KIB; Index++) {
    PropFillValue[Index] = 'X';
  }

  PropFillValue[KIB]   = '\0';
  MsgPropFillMP.Length = strlen(MsgPropFillMP.Value);

/********************************************************************************
**
** Check for queue manager and queue arguments.
**
** RunMode = 0: Broken and uses MQMD MsgDesc with MQPUT1
** RunMode = 1: Broken and uses NULL MsgDesc with MQPUT1
** RunMode = 2: Fixed  and uses NULL MsgDesc with MQPUT1
**
********************************************************************************/
  if (argc < 4) {
    printf("\nUSAGE: putwmp QMgrName QueueName RunMode [MaxMsgSize] [ReadMaxBytes] [MsgLengthMod]\n");
    exit(1);
  }

  QMgrName    = argv[1];
  QueName     = argv[2];
  QMgrNameLen = strlen(argv[1]);
  QueNameLen  = strlen(argv[2]);

  if (QMgrNameLen > MQ_Q_MGR_NAME_LENGTH) {
    printf("\nQueue manager \"%s\": name length = [%d] bytes, cannot exceed [%d] bytes. Try again.\n\n",
      QMgrName, QMgrNameLen, MQ_Q_MGR_NAME_LENGTH);
    exit(1);
  }
  if (QueNameLen > MQ_Q_NAME_LENGTH) {
    printf("\nQueue \"%s\": name length = [%d] bytes, cannot exceed [%d] bytes. Try again.\n\n",
      QueName, QueNameLen, MQ_Q_MGR_NAME_LENGTH);
    exit(1);
  }

  if (argc > 3) {                                               // if: arg count => the 3 required
    for (argCount = 3; argCount < argc; argCount++) {           // for: process run mode and optionally provided numeric values
      argvLen = strlen(argv[argCount]);

      for (Index = 0; Index < argvLen; Index++) {               // make sure value provided is numeric
        if (!isdigit(argv[argCount][Index])) break;
      }

      if (Index < argvLen) {
        if (argCount == 3) {
          printf("\nRunMode value must be numeric. Try again.\n\n");
        }
        else {
          printf("\nOptionally provided value #%d must be numeric. Try again.\n\n", (argCount - 4));
        }
        exit(1);
      }
      else {
        optArgs[argCount - 3] = atoi(argv[argCount]);
      }
    }                                                           // for: process optionally provided numeric values

    if (optArgs[0] < 3) {
      RunMode = optArgs[0];
    }
    else {
      printf("\nRunmode = [%d] out of range. Value must be 0, 1 or 2. Try again.\n\n", RunMode);
      exit(1);
    }

    if (optArgs[1] > 0) {
      MaxMsgSize = optArgs[1];
      printf("\nOptional max permitted message length of [%d] bytes will be used.\n", MaxMsgSize);
    }

    if (optArgs[2] > 0) {
      ReadMaxBytes = optArgs[2];
      printf("\nOptional max number of [%d] bytes will be read from text file.\n", ReadMaxBytes);
    }

    if (optArgs[3] != 0) {
      MsgLenMod = abs(optArgs[3]);
      printf("\nOptional message length modifier of [%d] bytes will be applied to the message length.\n",
        MsgLenMod);
    }
  }                                                             // end if: arg count > the 4 required

  if (RunMode < 2) {
    printf("\nWelcome to the 'broken' run mode of Whack-a-XmitQueue using \"%s\" for the MQPUT MsgDesc parameter.\n",
      (RunMode == 0) ? "MQMD" : "NULL");
  }
  else {
    printf("\nWelcome to the 'fixed' run mode of Whack-a-XmitQueue.\n");
  }
  printf("\nInitial size of all %d message property data structures is [%d] bytes.\n", MsgPropCount, MsgPropsInitSize);

  memcpy(QueObjDesc.ObjectName, QueName, (size_t) QueNameLen);
  MsgBufSize = 0;
  srand((unsigned)time(NULL));

/********************************************************************************
**
** Set MQMD used for large messages that will be segmented.
**
** (MQBYTE24) Root.MQMD.GroupId  MQGI_NONE { 'X000000000000000000000000' }
** (MQLONG)   Root.MQMD.MsgFlags MQMF_SEGMENTATION_ALLOWED { 1 }
**
********************************************************************************/
  memcpy(MsgHdrDesc.GroupId, MQGI_NONE, (size_t) MQ_GROUP_ID_LENGTH);

  MsgHdrDesc.MsgFlags    |=  MQMF_SEGMENTATION_ALLOWED;
  MsgHdrDesc.Persistence  =  MQPER_NOT_PERSISTENT;

  if (RunMode > 1) {                                            // if: RunMode = 'fixed'
    memcpy(MsgHdrDesc.Format, MQFMT_STRING, (size_t) MQ_FORMAT_LENGTH);
    MsgHdrDesc.MsgFlags |= MQMF_LAST_MSG_IN_GROUP;
  }

/********************************************************************************
**
** Connect to queue manager.
**
********************************************************************************/
  MQCONN (QMgrName,                    // queue manager name
          &QMgrHandle,                 // connection handle
          &CompCode,                   // completion code
          &Reason);                    // reason code

  if (CompCode == MQCC_FAILED) {
    printf("\nMQCONN failed with error [%d].\\nn", Reason);
    exit((int) Reason);
  }
  printf("\nConnect to queue manager \"%s\" succeeded.\n", QMgrName);

/********************************************************************************
** Create the message handle.
**
** NOTE: DO NOT set "PutMsgOpts.NewMsgHandle = MsgHandle".  Doing so  will result
**       in any MDE defined by the QMgr for auto-segmentation of a large message
**       to continue to be embedded in the message data payload.
**
** NOTE: The put queue is an alias queue that resolves to a non-local alias
**       cluster queue and as such cannot be queried for max message size.
**
********************************************************************************/
  MQCRTMH (QMgrHandle,                 // connection handle
           &CrtMsgHndOpts,             // create message handle options
           &MsgHandle,                 // message handle
           &CompCode,                  // completion code
           &Reason);                   // reason code

  if (Reason != MQRC_NONE) {
    printf("\nCreate message handle failed with error [%d].\n\n", Reason);
    exit((int) Reason);
  }
  else {
    printf("\nCreate message handle succeeded.\n");
    PutMsgOpts.NewMsgHandle       = MQHM_NONE;
    PutMsgOpts.OriginalMsgHandle  = MsgHandle;
  }

/********************************************************************************
**
**  Set message property: PropertyFiller
**
********************************************************************************/
  MQSETMP (QMgrHandle,                 // connection handle
           MsgHandle,                  // message handle
           &SetMsgPrpOpts,             // message property set options
           &MsgPropFillMP.Name,        // property name to be set
           &MsgPrpDesc,                // property descriptor
           MsgPrpTyp,                  // property type
           MsgPropFillMP.Length,       // size of the property value to be set
           MsgPropFillMP.Value,        // property value
           &CompCode,                  // completion code
           &Reason);                   // reason code

  if (Reason != MQRC_NONE) {
    printf("\nSet message property: name = \"%s\" failed with error [%d].\n\n",
      (char *) MsgPropFillMP.Name.VSPtr, Reason);
    exit((int) Reason);
  }
  else {
    MsgPropsInitSize += MsgPropFillMP.Length;

    printf("\nSet message property: name = \"%s\", length = [%d] bytes succeeded.\n",
      (char *) MsgPropFillMP.Name.VSPtr, MsgPropFillMP.Length);
    printf("\nCurrent size of all %d properties for THIS message is now [%d] bytes.\n",
      MsgPropCount, MsgPropsInitSize);
  }

/********************************************************************************
**
** Get file name from user input. Read file into message buffer.
**
********************************************************************************/
  while (1) {                                                   // while: someone wants to enter file names
    printf("\n\nEnter text file name or 'Q' to quit: ");

    fgets(TextFileName, sizeof(TextFileName), stdin);
    PtrHolder = strchr(TextFileName, '\n');                     // fgets includes NL which we must overwrite with null
    *PtrHolder = '\0';
    SysError = 0;

    if (toupper(TextFileName[0]) == 'Q') {                      // if: it's time to quit

/********************************************************************************
**
** Delete the message handle and close the queue manager connection.
**
********************************************************************************/
      MQDLTMH (QMgrHandle,             // connection handle
               &MsgHandle,             // message handle
               &DltMsgHndOpts,         // delete message handle options
               &CompCode,              // completion code
               &Reason);               // reason code

      if (CompCode == MQCC_FAILED) {
        printf("\nDelete message handle failed with error [%d].\n\n", Reason);
      }
      else {
        printf("\nMessage handle delete succeeded.\n");
      }

      MQDISC (&QMgrHandle,
              &CompCode,
              &Reason);

      if (CompCode == MQCC_FAILED) {
        printf("\nDisconnect from queue manager \"%s\" failed with error [%d].\n\n", QMgrName, Reason);
      }
      else {
        printf("\nDisconnect from queue manager \"%s\" succeeded.\n", QMgrName);
      }

      printf("\nCome back soon. Bye!\n\n");
      exit(0);
    }

/********************************************************************************
**
** Open the text file for read.
**
********************************************************************************/
    if (((Error = fopen_s(&fsTextFile, TextFileName, "rb")) != 0) || (fsTextFile == NULL)) {  // if: file open failed
      SysError = GetLastError();
      printf("\nOpen text file \"%s\" for read failed with error [%d], syserror [%d].\n",
        TextFileName, (int)Error, SysError);
      continue;
    }
    else {                                                      // else: file open succeeeded
//      printf("\nOpen text file \"%s\" for read succeeded.\n", TextFileName);

/********************************************************************************
**
** Get and set the size of the text file.
**
********************************************************************************/
      if ((Result = _fstat(_fileno(fsTextFile), &FileStatBuf)) != 0) {
        printf("\nStat of text file \"%s\" failed with return = [%d], error [%d].\n", TextFileName, Result, errno);
        continue;
      }

      TextFileSize = (MQLONG) FileStatBuf.st_size;

/********************************************************************************
**
** We only want the short print file name for a message property value.
**
********************************************************************************/
      if ((PtrFileName = strrchr(TextFileName, '\\')) != NULL) {
        *(PtrFileName++);
      }
      else {
        PtrFileName = TextFileName;
      }

      strcpy_s(FileNameValue, sizeof(FileNameValue), PtrFileName);
      FileNameMP.Length = (MQLONG) strlen(FileNameValue);

      if (TextFileSize > MaxFileSize) {                         // if: the file is too large, ask for the next file
        printf("\nText file \"%s\", size = [%d] bytes exceeds the permitted max size of [%d] bytes. Try again.\n\n",
          FileNameValue, TextFileSize, MaxFileSize);
        continue;
      }
      else {
        printf("\nText file name = \"%s\", file name length = [%d] bytes, file size = [%d].\n",
          FileNameMP.Value, FileNameMP.Length, TextFileSize);
      }

/********************************************************************************
**
**  Set message property: FileName
**
********************************************************************************/
      MQSETMP (QMgrHandle,             // connection handle
               MsgHandle,              // message handle
               &SetMsgPrpOpts,         // set message property options
               &FileNameMP.Name,       // property name to be set
               &MsgPrpDesc,            // property descriptor
               MsgPrpTyp,              // property type
               FileNameMP.Length,      // size of the property value to be set
               FileNameMP.Value,       // property value
               &CompCode,              // completion code
               &Reason);               // reason code

      if (Reason != MQRC_NONE) {
        printf("\nSet message property: name = \"%s\" failed with error [%d].\n\n",
          (char *) FileNameMP.Name.VSPtr, Reason);
        exit((int) Reason);
      }
      else {
        MsgPropsTotalSize = MsgPropsInitSize + FileNameMP.Length;
        printf("\nSet message property: name = \"%s\", value = \"%s\", length = [%d] bytes succeeded.\n",
          (char *) FileNameMP.Name.VSPtr, FileNameMP.Value, FileNameMP.Length);
        printf("\nCurrent size of all %d properties for THIS message is now [%d] bytes.\n",
          MsgPropCount, MsgPropsTotalSize);
      }

      if (TextFileSize > (MQLONG) (0.91 * MsgBufSize)) {        // if: file size exceeds 91% of the message buffer size, allocate more memory
        OldMsgBufSize = MsgBufSize;
        MsgBufSize = (MQLONG) (TextFileSize * 1.1);

        printf("\n     Text file \"%s\", size = [%d] bytes exceeds the current message buffer size of [%d] bytes.\n",
          TextFileName, TextFileSize, OldMsgBufSize);
//        printf("\nAttempting to allocate [%d] bytes of memory for message buffer.\n", MsgBufSize);

        if (OldMsgBufSize == 0L) {                              // if: this is the initial memory allocation event
          if ((MsgBufPtr = (PMQBYTE) malloc(MsgBufSize * sizeof(MQBYTE))) == NULL) {     // if: allocate memory failed
            printf("\n     Attempt to allocate [%d] bytes of memory for message buffer failed with error [%d].\n\n",
              MsgBufSize, errno);
            exit(errno);
          }
        }
        else {                                                  // else: this is a memory reallocation event
          if ((MsgBufPtr = (PMQBYTE) realloc(MsgBufPtr, MsgBufSize * sizeof(MQBYTE))) == NULL) {   // if: reallocate memory failed
            printf("\n     Attempt to allocate [%d] bytes of memory for message buffer failed with error [%d].\n\n",
              MsgBufSize, errno);
            exit(errno);
          }
        }
        printf("\n     Allocate [%d] bytes of memory for message buffer succeeded.\n", MsgBufSize);
      }

/********************************************************************************
**
**  Read contents of FileName into message buffer.
**
********************************************************************************/
      while (NOT feof(fsTextFile)) {                            // while: there's data to read into the message buffer
        BytesRead = fread(MsgBufPtr,
                          sizeof(MQBYTE),
                          (ReadMaxBytes == 0) ? MsgBufSize : ReadMaxBytes,
                          fsTextFile);

        if ((SysError = ferror(fsTextFile)) ||                  // if: file read error
             (BytesRead < 1L)) {
          if (fsTextFile != NULL) clearerr(fsTextFile);

          printf("\nRead text file \"%s\" failed with error [%d] ([%d] of [%d] bytes read).  \
Unable to process text file.\n\n", TextFileName, SysError, BytesRead, TextFileSize);
          exit((int) SysError);
        }
        else {
//          printf("\nRead [%d] bytes from text file \"%s\" succeeded.\n", BytesRead, TextFileName);
          if (ReadMaxBytes > 0) break;
        }
      }                                                         // end while: there's data to read into the message buffer

      MsgLen = BytesRead - MsgLenMod;                           // message length is the # of bytes read - message length modifier

      if ((fclose(fsTextFile)) != 0) {                          // close text file
        SysError = GetLastError();
        if (fsTextFile != NULL) clearerr(fsTextFile);

        printf("\nClose text file \"%s\" failed with error [%d].\n", TextFileName, SysError);
      }

      printf("\n     Text file \"%s\", size [%d] bytes: Read [%d] bytes into message buffer succeeded.\n",
        TextFileName, TextFileSize, MsgLen);

      sprintf_s(MsgLengthValue,                                 // set MsgLengthMP message property value and size
                sizeof(MsgLengthValue), "%08d", MsgLen);
      MsgLengthMP.Length = strlen(MsgLengthValue);

      if ((RunMode > 1) &&                                      // if: running 'fixed' mode, compute true message length
          (MsgLen > MaxMsgSize - (MsgPropsTotalSize + CLEARMPDS + CLEARXMIT))) {
        TrueMsgLen = MsgLen + MsgPropsTotalSize + CLEARMPDS + CLEARXMIT;

        printf("\nComputed effective message length = [%d] bytes.\n", TrueMsgLen);
        printf("   Message properties length = [%d] bytes; CLEARXMIT = [%d] bytes; CLEARMPDS = [%d] bytes.\n",
          MsgPropsTotalSize, CLEARXMIT, CLEARMPDS);
      }
      else {
        TrueMsgLen = MsgLen;
      }
    }                                                           // end else: file open succeeded

/********************************************************************************
**
** Set message property: MessageLength
**
********************************************************************************/
    MQSETMP (QMgrHandle,               // connection handle
             MsgHandle,                // message handle
             &SetMsgPrpOpts,           // set message property options
             &MsgLengthMP.Name,        // property name to be set
             &MsgPrpDesc,              // property descriptor
             MsgPrpTyp,                // property type
             MsgLengthMP.Length,       // size of the property value to be set
             MsgLengthMP.Value,        // property value
             &CompCode,                // completion code
             &Reason);                 // reason code

    if (Reason != MQRC_NONE) {
      printf("\nSet message property: name = \"%s\" failed with error [%d].\n\n",
        (char *) MsgLengthMP.Name.VSPtr, Reason);
      exit((int) Reason);
    }
    else {
      MsgPropsTotalSize += MsgLengthMP.Length;
      printf("\nSet message property: name = \"%s\", value = \"%s\", length = [%d] bytes succeeded.\n",
        (char *) MsgLengthMP.Name.VSPtr, MsgLengthMP.Value, MsgLengthMP.Length);
      printf("\nCurrent size of all %d properties for THIS message is now [%d] bytes.\n",
        MsgPropCount, MsgPropsTotalSize);
    }

/********************************************************************************
**
** Setup for putting a large message for auto-segmentation.
**
** If the message length (= size of text file) plus the total length of all
** message properties exceeds 4MiB, then the message must be segmented.
**
********************************************************************************/
    if ((RunMode > 1) && (TrueMsgLen > MaxMsgSize)) {           // if: we're running in 'fixed' mode and size of message > max message size
      printf("\n     True message size of [%d] bytes exceeds maximum computed message size of [%d] bytes.  \
Message will be segmented.\n", TrueMsgLen, MaxMsgSize);

      BigMsgFlag = 1;                                           // big message switch

      GrpMsgNum = 2;                                            // big message group message number (2 logical messages in group)

      sprintf_s(RandID, sizeof(RandID), "%08d", rand());        // build the group ID randomizer
      sprintf_s(MsgGrpIdValue, sizeof(MsgGrpIdValue), "%s%s", GRPIDHEADER, RandID);

      MsgGrpIdMP.Length = strlen(MsgGrpIdValue);

      printf("\n     Eight digit random ID [%s] generated for message group ID.\n", RandID);
      printf("\n     Segmented message group ID = \"%s\", length = [%d] bytes.\n",
        MsgGrpIdValue, MsgGrpIdMP.Length);
    }                                                           // end if: running 'fixed' mode && message size > max message size

    else {                                                      // else: message size < max message size
      BigMsgFlag = 0;

      if (MsgGrpIdMP.Length > 0) {                              // if: the last message was a large message then
        memcpy(MsgGrpIdValue, MQGI_NONE, (size_t) MQ_GROUP_ID_LENGTH);
        MsgGrpIdMP.Length = 0;                                  //     reset the MessageGroupId MP back to null
      }
    }

/********************************************************************************
**
**  Set message property: MessageGroupId
**
********************************************************************************/
    MQSETMP (QMgrHandle,               // connection handle
             MsgHandle,                // message handle
             &SetMsgPrpOpts,           // message property set options
             &MsgGrpIdMP.Name,         // property name to be set
             &MsgPrpDesc,              // property descriptor
             MsgPrpTyp,                // property type
             MsgGrpIdMP.Length,        // size of the property value to be set
             MsgGrpIdMP.Value,         // property value
             &CompCode,                // completion code
             &Reason);                 // reason code

    if (Reason != MQRC_NONE) {
      printf("\nSet message property: name = \"%s\" failed with error [%d].\n\n",
        (char *) MsgGrpIdMP.Name.VSPtr, Reason);
      exit((int) Reason);
    }
    else {
      MsgPropsTotalSize += MsgGrpIdMP.Length;
      printf("\nSet message property: name = \"%s\", value = \"%s\", length = [%d] bytes succeeded.\n",
        (char *) MsgGrpIdMP.Name.VSPtr, MsgGrpIdMP.Value, MsgGrpIdMP.Length);
      printf("\nFinal size of all %d properties for THIS message is [%d] bytes.\n",
        MsgPropCount, MsgPropsTotalSize);
    }
;
/********************************************************************************
**
** Put the message.
**
** BigMsgFlag is either 0 or 1; GrpMsgNum is either 1 or 2
**
** A non-segmented message group always has one message in it. A segmented message
** group always has two messages in it.
********************************************************************************/
    while (GrpMsgNum > 0) {                                     // while: there's a message in the group
      switch (RunMode) {
        case 0: MQPutSwitch = -1;                               // if: we're running in 'broken' mode using MQMD MsgDesc
          break;
        case 1: MQPutSwitch = 0;                                // if: we're running in 'broken' mode using NULL MsgDesc
          break;
        case 2: MQPutSwitch = GrpMsgNum * BigMsgFlag;           // if: we're running in 'fixed' mode

                if (MQPutSwitch == 1) {                         // if: this is the 2nd message of a big message group
                  PutMsgOpts.OriginalMsgHandle = MQHM_NONE;     //     then set PMO.OriginalMsgHandle = MQHM_NONE
                                                                //     and set the MQMD.GroupId to the generated value
                  memcpy(MsgHdrDesc.GroupId, MsgGrpIdValue, (size_t) MsgGrpIdMP.Length);
                }
          break;
      }

      MQPUT1 (QMgrHandle,                                       // connection handle
              &QueObjDesc,                                      // object descriptor
              (abs(MQPutSwitch) == 1) ? &MsgHdrDesc :           // message header
                                        (PMQMD) NULL,
              &PutMsgOpts,                                      // put message options
              (MQPutSwitch == 2) ?  1  : MsgLen,                // message length
              (MQPutSwitch == 2) ? "1" : MsgBufPtr,             // message buffer
              &CompCode,                                        // completion code
              &Reason);                                         // reason code

      if (Reason == MQRC_NONE) {
        if (MQPutSwitch > 0) {
          printf("\nSegmented message group message #%d, put switch #%d MQPUT to queue \"%s\" succeeded.\n",
            (GrpMsgNum == 2) ? 1 : 2, MQPutSwitch, QueObjDesc.ObjectName);
        }
        else {
          printf("\nMessage #1 MQPUT to queue \"%s\" succeeded.\n", QueObjDesc.ObjectName);
        }
      }
      else {
        printf("\nMessage MQPUT to queue \"%s\" failed with error [%d].\n", QueObjDesc.ObjectName, Reason);
        exit((int) Reason);
      }

      if (MQPutSwitch == 1) {                                   // if: this was the 2nd message of a big message group
        PutMsgOpts.OriginalMsgHandle = MsgHandle;               //     then reset PMO.OriginalMsgHandle = MsgHandle
                                                                //     and reset the MQMD.GroupId back to MQGI_NONE
        memcpy(MsgHdrDesc.GroupId, MQGI_NONE, (size_t) MQ_GROUP_ID_LENGTH);
      }
      GrpMsgNum--;                                              // decrement the message number in the group
    }                                                           // end while: there's a message in the group
    GrpMsgNum = 1;                                              // reset GrpMsgNum back to 1
  }                                                             // end while: someone wants to enter file names
}
